/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.ExecutionException;

public class GFServer extends Thread {

  private final String         prefix;
  private final String         twoBitFile;
  private int                  portNum;
  private final ProcessBuilder pb;
  private Process              proc;
  private boolean              debug;

  // TO DO: Need to fail nice for non-unix by checking for lsof/pgrep
  public GFServer(String prefix, String twoBitFile, boolean debug) throws InterruptedException, IOException,
      ExecutionException {
    this.prefix = prefix;
    this.twoBitFile = twoBitFile;
    this.debug = debug;
    portNum = getNextAvailablePort();
    if (portNum == -1) {
      throw new InterruptedException("Could not locate a port number to use!");
    }
    pb = new ProcessBuilder("gfServer", "start", "localhost", "" + portNum, twoBitFile, "-canStop", "-log=" + prefix
        + ".gfServer.log");
  }

  private int getNextAvailablePort() throws InterruptedException, IOException, ExecutionException {
    // String portCheck = RunCommand.run(new ProcessBuilder("lsof", "-i",
    // "@0.0.0.0:" + port));
    // while (portCheck.length() != 0) {
    // port++;
    // portCheck = RunCommand.run(new ProcessBuilder("lsof", "-i", "@0.0.0.0:" +
    // port));
    // }
    String processCheck = RunSystemCommand.run(new ProcessBuilder("pgrep", "-l", "-f", "gfServer"), debug);
    for (int i = 49152; i <= 65535; i++) {
      String portCheck = RunSystemCommand.run(new ProcessBuilder("lsof", "-i", "@0.0.0.0:" + i), debug);
      if ((portCheck.length() != 0) || processCheck.contains("" + i)) {
        continue;
      } else {
        return i;
      }
    }
    return -1;
    // for (int i = 49152; i <= 65535; i++) {
    // ServerSocket serverSocket = null;
    // DatagramSocket datagramSocket = null;
    // try {
    // serverSocket = new ServerSocket(i);
    // serverSocket.setReuseAddress(true);
    // datagramSocket = new DatagramSocket(i);
    // datagramSocket.setReuseAddress(true);
    // return i;
    // } catch (IOException e) {
    // continue;
    // } finally {
    // if (datagramSocket != null) {
    // datagramSocket.close();
    // }
    // if (serverSocket != null) {
    // try {
    // serverSocket.close();
    // } catch (IOException e) {
    // // Can never happen
    // throw new InterruptedException("Could not locate a port number to use!");
    // }
    // }
    // }
    // }
    // return -1;
  }

  // private boolean isPortFree(int port) throws InterruptedException,
  // IOException, ExecutionException {
  // String portCheck = RunCommand.run(new ProcessBuilder("pgrep", "-l", "-f",
  // "gfServer"));
  // if (portCheck.contains("" + port)) {
  // return false;
  // } else {
  // return true;
  // }
  // }

  @Override
  public void run() {
    try {
      pb.redirectErrorStream(true);
      proc = pb.start();
      InputStream in = proc.getInputStream();
      FileOutputStream out = new FileOutputStream(prefix + ".gfServer.out");
      byte[] buf = new byte[1024];
      int len;
      while ((len = in.read(buf)) > 0) {
        out.write(buf, 0, len);
        out.flush();
      }
      in.close();
      out.close();
      proc.waitFor();
    } catch (InterruptedException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  public void waitForReady() throws InterruptedException, IOException, ExecutionException {
    while (proc == null) {
      sleep(1000);
    }
    String serverCheck = RunSystemCommand
        .run(new ProcessBuilder("gfServer", "files", "localhost", "" + portNum), debug);
    while (!serverCheck.trim().equals(twoBitFile)) {
      sleep(1000);
      serverCheck = RunSystemCommand.run(new ProcessBuilder("gfServer", "files", "localhost", "" + portNum), debug);
    }
  }

  public int getPortNum() {
    return portNum;
  }

  public String getTwoBitFile() {
    return twoBitFile;
  }

  public void close() {
    if (proc != null) {
      proc.destroy();
    }
  }
}
